/*
 * Copyright (c) 2002 Bang Jun-Young
 * All rights reserved.
 */

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "rtld.h"
#include "debug.h"

#define RELCACHE_DIR	"/var/db/relcache/"

int
_rtld_create_relcache(Obj_Entry *first)
{
	Obj_Entry *obj;
	int fd;
	size_t nbytes = 0;
	int offset = _rtld_pagesz;
	int nents = 1;
	RelCache_Entry *rcache;
	char *buf;
	char path[64];

	if (strlen(_rtld_objself.cksum) == 0) {
		dbg(("rtld md5 checksum not found"));
		goto bad;
	}
	for (obj = first; obj != NULL; obj = obj->next)
		if (!obj->cksumfound || *obj->cksum == '\0') {
			dbg(("shobj md5 checksum not found"));
			goto bad;
		}
	for (obj = first; obj != NULL; obj = obj->next)
		nents++;
	nbytes = round_up(nents * sizeof(RelCache_Entry) + 4);
	/* XXX this limitation should be gone. */
	if (nbytes > _rtld_pagesz)
		goto bad;

	buf = xcalloc(nbytes);
	memcpy(buf, &nents, 4);
	rcache = (RelCache_Entry *)(buf + 4);

	for (obj = first; obj != NULL; obj = obj->next) {
		memcpy(rcache, obj->cksum, 32);
		rcache->base = obj->database;
		rcache->offset = offset;
		rcache->size = obj->datasize;
		dbg(("database=%p, size=0x%x, offset=0x%x",
		    obj->database, obj->datasize, offset));
		offset += obj->datasize;
		rcache++;
	}

	memcpy(rcache, _rtld_objself.cksum, 32);
	rcache->base = 0;
	rcache->offset = 0;
	rcache->size = 0;

	strcpy(path, RELCACHE_DIR);
	strcat(path, first->cksum);
	fd = open(path, O_CREAT | O_TRUNC | O_WRONLY, 0644);
	if (fd < 0) {
		xwarn("%s", path);
		return -1;
	}

	write(fd, buf, nbytes);

	for (obj = first; obj != NULL; obj = obj->next)
		write(fd, obj->database, obj->datasize);

	close(fd);
	free(buf);
	return 0;
bad:
	close(fd);
	return -1;
}

bool
_rtld_load_relcache(Obj_Entry *first)
{
	caddr_t cachebase;
	int nents, nobjs = 1;
	Obj_Entry *obj;
	char path[64];
	int fd, i;
	RelCache_Entry *rcache;

	strcpy(path, RELCACHE_DIR);
	strcat(path, first->cksum);
	fd = open(path, O_RDONLY);
	if (fd < 0)
		return false;

	cachebase = mmap(0, _rtld_pagesz, PROT_READ, MAP_FILE | MAP_PRIVATE,
	    fd, 0);
	if (cachebase == MAP_FAILED)
		goto bad2;
	nents = (int)*cachebase;
	if (!_rtld_objself.cksumfound || *_rtld_objself.cksum == '\0') {
		dbg(("rtld md5 checksum not found"));
		goto bad;
	}
	for (obj = first; obj != NULL; obj = obj->next) {
		if (!obj->cksumfound || *obj->cksum == '\0')
			goto bad;
		nobjs++;
	}
	if (nents != nobjs) {
		dbg(("nents %d != nobjs %d", nents, nobjs));
		goto bad;
	}

	/* XXX this limitation should be gone. */
	if (nents * sizeof(RelCache_Entry) + 4 > _rtld_pagesz) {
		dbg(("header too big: %d > %d",
		    nents * sizeof(RelCache_Entry) + 4, _rtld_pagesz));
		goto bad;
	}

	rcache = (RelCache_Entry *)(cachebase + 4);
	if (strcmp(rcache[nents - 1].md5, _rtld_objself.cksum) != 0) {
		dbg(("rtld md5 checksum mismatch"));
		goto bad;
	}
	nobjs--;

	for (obj = first; obj != NULL; obj = obj->next) {
		for (i = 0; i < nents - 1; i++)
			if (strcmp(rcache[i].md5, obj->cksum) == 0 &&
			    rcache[i].base == obj->database) {
				nobjs--;
				obj->relcache = &rcache[i];
			}

		/* Set some sanity-checking numbers in the Obj_Entry. */
		obj->magic = RTLD_MAGIC;
		obj->version = RTLD_VERSION;

		/* Fill in the dynamic linker entry points. */
		obj->dlopen = _rtld_dlopen;
		obj->dlsym = _rtld_dlsym;
		obj->dlerror = _rtld_dlerror;
		obj->dlclose = _rtld_dlclose;
		obj->dladdr = _rtld_dladdr;
	}

	if (nobjs != 0) {
		dbg(("shobj md5 checksum mismatch"));
		goto bad;
	}

	for (obj = first; obj != NULL; obj = obj->next) {
		if (mmap(obj->database, obj->relcache->size,
		    obj->dataflags, MAP_FILE | MAP_PRIVATE | MAP_FIXED, fd,
		    obj->relcache->offset) == MAP_FAILED) {
			xwarn("mmap of data failed");
			goto bad;
		}
	}

	munmap(cachebase, _rtld_pagesz);
	close(fd);
	return true;
bad:
	munmap(cachebase, _rtld_pagesz);
bad2:
	close(fd);
	return false;
}
